"""Helpers used to serialize/deserialize Atop statistics directly from log files.

The basic layout of an Atop log file can be described as the following:
1. Raw Header:
    - Single instance at the beginning of the file.
    - Contains information about all the following raw records.
2. Raw Record:
    - Single instance repeated in a loop with remaining stats until end of file.
    - Contains information about the *Stat structs that immediately follow it.
3. Raw SStat:
    - Single instance repeated in a loop with remaining stats until end of file.
    - Always found directly after a Record.
    - Always found directly before a TStat array.
    - Contains statistics about overall system activity.
4. Raw TStats:
    - Array of TStat instances repeated in a loop with remaining stats until end of file.
    - Always found directly after a SStat.
    - Always found directly before the next Record (if <= 2.11) or before a CStat array (if 2.11+).
    - Contains statistics about every task/process on the system.
4. Raw CStats/CGroups:
    - Array of CStat/CGroup instances repeated in a loop with other stats until end of file.
    - Always found directly after a TStat.
    - Always found directly before the next Record if not the end of file.
    - Contains statistics about every CGroup on the system.
"""

from __future__ import annotations

import ctypes
import io
import zlib
from typing import Union

from atoparser.structs import atop_1_26
from atoparser.structs import atop_2_3
from atoparser.structs import atop_2_4
from atoparser.structs import atop_2_5
from atoparser.structs import atop_2_6
from atoparser.structs import atop_2_7
from atoparser.structs import atop_2_8
from atoparser.structs import atop_2_9
from atoparser.structs import atop_2_10
from atoparser.structs import atop_2_11
from atoparser.structs import atop_2_12
from atoparser.structs.shared import pid_t

_VERSIONS = [
    atop_1_26,
    atop_2_3,
    atop_2_4,
    atop_2_5,
    atop_2_6,
    atop_2_7,
    atop_2_8,
    atop_2_9,
    atop_2_10,
    atop_2_11,
    atop_2_12,
]
_CSTAT_VERSIONS = _VERSIONS[9:]
# Fallback to latest if there is no custom class provided to attempt backwards compatibility.
_DEFAULT_VERSION = _VERSIONS[-1].Header.supported_version
_HEADER_BY_VERSION: dict[str, type[Header]] = {module.Header.supported_version: module.Header for module in _VERSIONS}

Header = Union[tuple(header for header in _HEADER_BY_VERSION.values())]
Record = Union[tuple(header.Record for header in _HEADER_BY_VERSION.values())]
SStat = Union[tuple(header.SStat for header in _HEADER_BY_VERSION.values())]
TStat = Union[tuple(header.TStat for header in _HEADER_BY_VERSION.values())]  # pylint: disable=invalid-name
CStat = Union[tuple(header.CStat for header in _HEADER_BY_VERSION.values())]
CGChainer = Union[tuple(header.CGChainer for header in _HEADER_BY_VERSION.values())]


# Default Atop sample is once per minute, but it can be manually increased/decreased.
# Additionally, logs may not rollover as expected, combining multiple hours into 1 log.
# Limit the amount of read attempts to 1 per second for 1 day to ensure we do not get stuck in an infinite loop.
MAX_SAMPLES_PER_FILE = 86400

# Definition from rawlog.c
MAGIC = 0xFEEDBEEF


def generate_statistics(
    raw_file: io.BytesIO,
    header: Header = None,
    raise_on_truncation: bool = True,
    max_samples: int = MAX_SAMPLES_PER_FILE,
) -> tuple[Record, SStat, list[TStat], list[CGChainer]]:
    """Read statistics groups from an open Atop log file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        header: The header from the file containing metadata about records to read. If not provided, one will be read.
        raise_on_truncation: Raise compression exceptions after header is read. e.g. Software restarts
        max_samples: Maximum number of samples read from a file.

    Yields:
        The next record, sstat, tstat list, and cstat list statistic groups after reading in raw bytes to objects.
    """
    if header is None:
        # If a header was not provided, read up to the proper length and discard to ensure the correct starting offset.
        header = get_header(raw_file)

    try:
        header_version = header.semantic_version
        major, minor = header_version.split(".")[:2]
        major, minor = int(major), int(minor)
        for _ in range(max_samples):
            # Read the repeating structured information until the end of the file.
            # Atop log files consist of the following after the header, repeated until the end:
            # 1. Record: Metadata about statistics.
            # 2. SStats: System statistics.
            # 3. TStats: Task/process statistics.
            # 4. CGChain/CStats: CGroup statistics.
            record = get_record(raw_file, header)
            if record.scomplen <= 0:
                # Natural end-of-file, no further bytes were found to populate another record.
                break
            sstat = get_sstat(raw_file, header, record)
            tstats = get_tstat(raw_file, header, record)
            if major >= 2 and minor >= 11:
                cgroups = get_cstat(raw_file, header, record)
            else:
                cgroups = []
            yield record, sstat, tstats, cgroups
    except zlib.error:
        # End of readable data reached. This is common during software restarts.
        # All errors after the header are squashable errors, since that means the file is valid, but was not closed.
        if raise_on_truncation:
            raise


def get_header(raw_file: io.BytesIO, check_compatibility: bool = True) -> Header:
    """Get the raw file header from an open Atop file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        check_compatibility: Whether to enforce compatibility check against supported versions on header creation.

    Returns:
        The header at the beginning of an Atop file.

    Raises:
        ValueError if there are not enough bytes to read the header, or the bytes were invalid.
    """
    # Read the header directly into the struct, there is no padding to consume or add.
    # Use default Header as the baseline in order to check the version. It can be transferred without re-reading.
    header = _HEADER_BY_VERSION[_DEFAULT_VERSION]()
    raw_file.readinto(header)

    if header.magic != MAGIC:
        msg = f"File does not contain raw atop output (wrong magic number): {hex(header.magic)}"
        raise ValueError(msg)

    header_version = header.semantic_version
    if header_version != _DEFAULT_VERSION and header_version in _HEADER_BY_VERSION:
        # Header byte length is consistent across versions. Transfer the initial read into the versioned header.
        header = _HEADER_BY_VERSION[header_version].from_buffer(header)

    if check_compatibility:
        # Ensure all struct lengths match the lengths specified in the header. If not, we cannot read the file further.
        header.check_compatibility()

    return header


def get_record(
    raw_file: io.BytesIO,
    header: Header,
) -> Record:
    """Get the next raw record from an open Atop file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        header: The header from the file containing metadata about records to read.

    Returns:
        A single record representing the data before an SStat struct.

    Raises:
        ValueError if there are not enough bytes to read a single record.
    """
    record = header.Record()
    raw_file.readinto(record)
    return record


def get_sstat(
    raw_file: io.BytesIO,
    header: Header,
    record: Record,
) -> SStat:
    """Get the next raw sstat from an open Atop file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        header: The header from the file containing metadata about records to read.
        record: The preceding record containing metadata about the SStat to read.

    Returns:
        A single struct representing the data after a raw record, but before an array of TStat structs.

    Raises:
        ValueError if there are not enough bytes to read a single stat.
    """
    # Read the requested length instead of the length of the struct.
    # The data is compressed and must be decompressed before it will fill the struct.
    buffer = raw_file.read(record.scomplen)
    decompressed = zlib.decompress(buffer)
    sstat = header.SStat.from_buffer_copy(decompressed)
    return sstat


def get_tstat(
    raw_file: io.BytesIO,
    header: Header,
    record: Record,
) -> list[TStat]:
    """Get the next raw tstat array from an open Atop file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        header: The header from the file containing metadata about records to read.
        record: The preceding record containing metadata about the TStats to read.

    Returns:
        All TStat structs after a raw SStat, but before the next raw record.

    Raises:
        ValueError if there are not enough bytes to read a stat array.
    """
    # Read the requested length instead of the length of the struct.
    # The data is compressed and must be decompressed before it will fill the final list of structs.
    buffer = raw_file.read(record.pcomplen)
    decompressed = zlib.decompress(buffer)

    record_count = record.nlist if isinstance(record, atop_1_26.Record) else record.ndeviat
    tstatlen = header.tstatlen if header.major_version >= 2 and header.minor_version >= 3 else header.pstatlen
    tstats = []
    for index in range(record_count):
        # Reconstruct one TStat struct for every possible byte chunk, incrementing the offset each pass. For example:
        # First pass: 0 - 21650
        # Second pass: 21651 - 43300
        tstat = header.TStat.from_buffer_copy(decompressed[index * tstatlen : tstatlen * (index + 1)])
        tstats.append(tstat)
    return tstats


def get_cstat(
    raw_file: io.BytesIO,
    header: Header,
    record: Record,
) -> list[CGChainer]:
    """Get the next raw cstat array from an open Atop file.

    Args:
        raw_file: An open Atop file capable of reading as bytes.
        header: The header from the file containing metadata about records to read.
        record: The preceding record containing metadata about the CStats to read.

    Returns:
        All CGroup CStat structs and PID lists after a raw TStat, but before the next raw record.

    Raises:
        ValueError if there are not enough bytes to read a stat array.
    """
    # Read the requested length instead of the length of the struct.
    # The data is compressed and must be decompressed before it will fill the final list of structs.
    buffer_cstats = raw_file.read(record.ccomplen)
    decompressed_cstats = zlib.decompress(buffer_cstats)
    buffer_pidlist = raw_file.read(record.icomplen)
    decompressed_pidlist = zlib.decompress(buffer_pidlist)

    cgroups = []
    cstat_start = 0
    cstatlen = header.cstatlen
    pidlist_start = 0
    for _ in range(record.ncgroups):
        # Reconstruct one CStat struct and pidlist for every possible byte chunk, incrementing the offset each pass.
        # For example:
        # First pass: 0 - 21650
        # Second pass: 21651 - 43300
        # N.B. The variable length cgname is currently unsupported. In order to properly skip the remaining bytes,
        # Use the final structlen from the nested gen struct to update the starting point.
        cstat = header.CStat.from_buffer_copy(decompressed_cstats[cstat_start : cstat_start + cstatlen])
        cstat_start += cstat.gen.structlen

        pid_array = pid_t * cstat.gen.nprocs
        pid_array_size = ctypes.sizeof(pid_array)
        pidlist = pid_array.from_buffer_copy(decompressed_pidlist[pidlist_start : pidlist_start + pid_array_size])
        pidlist_start += pid_array_size

        cgroups.append(header.CGChainer(cstat, pidlist))
    return cgroups


def struct_to_dict(struct: ctypes.Structure) -> dict:
    """Convert C struct, and all nested structs, into a Python dictionary.

    Skips any "future" named fields since they are empty placeholders for potential future versions.

    Args:
        struct: C struct loaded from raw Atop file.

    Returns:
        C struct converted into a dictionary using the names of the struct's fields as keys.
    """
    struct_dict = {}
    for field in struct._fields_:  # pylint: disable=protected-access
        field_name = field[0]
        field_data = getattr(struct, field_name)
        if isinstance(field_data, ctypes.Structure):
            struct_dict[field_name] = struct_to_dict(field_data)
        elif "future" not in field_name:
            if isinstance(field_data, ctypes.Array):
                struct_dict[field_name] = []
                limiters = getattr(struct, "fields_limiters", {})
                limiter = limiters.get(field_name)
                if limiter:
                    field_data = field_data[: getattr(struct, limiter)]
                for sub_data in field_data:
                    if isinstance(sub_data, ctypes.Structure):
                        struct_dict[field_name].append(struct_to_dict(sub_data))
                    else:
                        struct_dict[field_name].append(sub_data)
            elif isinstance(field_data, bytes):
                struct_dict[field_name] = field_data.decode(errors="ignore")
            else:
                struct_dict[field_name] = field_data
    return struct_dict



"""Structs and definitions used serialize/deserialize Atop statistics directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h

Struct ordering matches the C source to help with comparisons.
If structs match exactly from a previous version, they are reused via aliasing.

See https://github.com/Atoptool/atop for more information and full details about each field.
Using schemas and structs from Atop 2.5.0.
"""

from atoparser.structs import atop_1_26
from atoparser.structs import atop_2_3
from atoparser.structs import atop_2_4

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name,too-few-public-methods

# Definitions from atop.h
ACCTACTIVE = 0x00000001
IOSTAT = 0x00000004
NETATOP = 0x00000010
NETATOPD = 0x00000020
DOCKSTAT = 0x00000040
GPUSTAT = 0x00000080

# Definitions from photoproc.h
PNAMLEN = 15
CMDLEN = 255

# Definitions from photosyst.h
MAXCPU = 2048
MAXDSK = 1024
MAXLVM = 2048
MAXMDD = 256
MAXINTF = 128
MAXCONTAINER = 128
MAXNFSMOUNT = 64
MAXIBPORT = 32
MAXGPU = 32
MAXGPUBUS = 12
MAXGPUTYPE = 12
MAXDKNAM = 32
MAXIBNAME = 12


Record = atop_2_3.Record


MemStat = atop_2_3.MemStat


FreqCnt = atop_1_26.FreqCnt


PerCPU = atop_2_3.PerCPU


CPUStat = atop_2_3.CPUStat


PerDSK = atop_2_3.PerDSK


DSKStat = atop_2_3.DSKStat


PerIntf = atop_2_3.PerIntf


IntfStat = atop_2_3.IntfStat


PerNFSMount = atop_2_3.PerNFSMount


Server = atop_2_3.Server


Client = atop_2_3.Client


NFSMounts = atop_2_3.NFSMounts


NFSStat = atop_2_3.NFSStat


PSI = atop_2_4.PSI


Pressure = atop_2_4.Pressure


PerContainer = atop_2_3.PerContainer


ContStat = atop_2_3.ContStat


WWWStat = atop_1_26.WWWStat


PerGPU = atop_2_4.PerGPU


GPUStat = atop_2_4.GPUStat


PerIFB = atop_2_4.PerIFB


IFBStat = atop_2_4.IFBStat


IPv4Stats = atop_1_26.IPv4Stats


ICMPv4Stats = atop_1_26.ICMPv4Stats


UDPv4Stats = atop_1_26.UDPv4Stats


TCPStats = atop_1_26.TCPStats


IPv6Stats = atop_1_26.IPv6Stats


ICMPv6Stats = atop_1_26.ICMPv6Stats


UDPv6Stats = atop_1_26.UDPv6Stats


NETStat = atop_1_26.NETStat


SStat = atop_2_4.SStat


GEN = atop_2_3.GEN


CPU = atop_1_26.CPU


DSK = atop_1_26.DSK


MEM = atop_2_3.MEM


NET = atop_2_3.NET


GPU = atop_2_4.GPU


TStat = atop_2_4.TStat


class Header(atop_2_3.Header):
    """Top level struct to describe information about the system running Atop and the log file itself."""

    supported_version = "2.5"
    Record = Record
    SStat = SStat
    TStat = TStat
    CStat = None
    CGChainer = None



"""Parser function used to convert Atop structs in 1.26 into "parseable" output based on the signature name."""

from atoparser import utils
from atoparser.structs import atop_1_26

# Disable the following pylint warnings to allow functions to match a consistent type across all parseables.
# This helps simplify calls allow dynamic function lookups to have consistent input arguments.
# pylint: disable=unused-argument,invalid-name


def parse_cpu(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'cpu' parseable representing per core usage."""
    for index, cpu in enumerate(sstat.cpu.cpu):
        if index >= sstat.cpu.nrcpu:
            # Core list contains 100 entries, but only up the count specified in the cpu stat are valid.
            break
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "ticks": header.hertz,
            "proc": index,
            "system": cpu.stime,
            "user": cpu.utime,
            "niced": cpu.ntime,
            "idle": cpu.itime,
            "wait": cpu.wtime,
            "irq": cpu.Itime,
            "softirq": cpu.Stime,
            "steal": cpu.steal,
            "guest": cpu.guest,
        }
        yield values


def parse_CPL(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'CPL' parseable representing system load."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "procs": sstat.cpu.nrcpu,
        "load_1": sstat.cpu.lavg1,
        "load_5": sstat.cpu.lavg5,
        "load_15": sstat.cpu.lavg15,
        "context_switches": sstat.cpu.csw,
        "interrupts": sstat.cpu.devint,
    }
    yield values


def parse_CPU(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Statistics for Atop 'CPU' parseable representing usage across cores combined."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "ticks": header.hertz,
        "procs": sstat.cpu.nrcpu,
        "system": sstat.cpu.all.stime,
        "user": sstat.cpu.all.utime,
        "niced": sstat.cpu.all.ntime,
        "idle": sstat.cpu.all.itime,
        "wait": sstat.cpu.all.wtime,
        "irq": sstat.cpu.all.Itime,
        "softirq": sstat.cpu.all.Stime,
        "steal": sstat.cpu.all.steal,
        "guest": sstat.cpu.all.guest,
    }
    yield values


def parse_DSK(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'DSK' parseable representing disk/drive usage."""
    for disk in sstat.dsk.dsk:
        if not disk.name:
            # Disk list contains 256 entries, but only up the first empty name are valid.
            break
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "name": disk.name.decode(),
            "io_ms": disk.io_ms,
            "reads": disk.nread,
            "read_sectors": disk.nrsect,
            "writes": disk.nwrite,
            "write_sectors": disk.nwsect,
        }
        yield values


def parse_LVM(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'LVM' parseable representing logical volume usage."""
    for lvm in sstat.dsk.lvm:
        if not lvm.name:
            # LVM list contains 256 entries, but only up the first empty name are valid.
            break
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "name": lvm.name.decode(),
            "io_ms": lvm.io_ms,
            "reads": lvm.nread,
            "read_sectors": lvm.nrsect,
            "writes": lvm.nwrite,
            "write_sectors": lvm.nwsect,
        }
        yield values


def parse_MDD(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'MDD' parseable representing multiple device drive usage."""
    for mdd in sstat.dsk.mdd:
        if not mdd.name:
            # MDD list contains 256 entries, but only up the first empty name are valid.
            break
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "name": mdd.name.decode(),
            "io_ms": mdd.io_ms,
            "reads": mdd.nread,
            "read_sectors": mdd.nrsect,
            "writes": mdd.nwrite,
            "write_sectors": mdd.nwsect,
        }
        yield values


def parse_MEM(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'MEM' parseable representing memory usage."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "page_size": header.pagesize,
        "phys_mem": sstat.mem.physmem,
        "free_mem": sstat.mem.freemem,
        "page_cache": sstat.mem.cachemem,
        "buffer_cache": sstat.mem.buffermem,
        "slab": sstat.mem.slabmem,
        "dirty_pages": sstat.mem.cachedrt,
    }
    yield values


def parse_NETL(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'NET' parseable representing network usage on lower interfaces."""
    for interface in sstat.intf.intf:
        if not interface.name:
            # Interface list contains 32 entries, but only up the first empty name are valid.
            break
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "name": interface.name.decode(),
            "pkt_received": interface.rpack,
            "byte_received": interface.rbyte,
            "pkt_transmitted": interface.spack,
            "bytes_transmitted": interface.sbyte,
            "speed": interface.speed,
            "duplex": int.from_bytes(interface.duplex, byteorder="big"),
        }
        yield values


def parse_NETU(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'NET' parseable representing network usage on upper interfaces."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "name": "upper",
        "tcp_pkt_received": sstat.net.tcp.InSegs,
        "tcp_pkt_transmitted": sstat.net.tcp.OutSegs,
        "udp_pkt_received": sstat.net.udpv4.InDatagrams + sstat.net.udpv6.Udp6InDatagrams,
        "udp_pkt_transmitted": sstat.net.udpv4.OutDatagrams + sstat.net.udpv6.Udp6OutDatagrams,
        "ip_pkt_received": sstat.net.ipv4.InReceives + sstat.net.ipv6.Ip6InReceives,
        "ip_pkt_transmitted": sstat.net.ipv4.OutRequests + sstat.net.ipv6.Ip6OutRequests,
        "ip_pkt_delivered": sstat.net.ipv4.InDelivers + sstat.net.ipv6.Ip6InDelivers,
        "ip_pkt_forwarded": sstat.net.ipv4.ForwDatagrams + sstat.net.ipv6.Ip6OutForwDatagrams,
    }
    yield values


def parse_PAG(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PAG' parseable representing paging space usage."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "page_size": header.pagesize,
        "page_scans": sstat.mem.pgscans,
        "alloc_stalls": sstat.mem.allocstall,
        "swapins": sstat.mem.swins,
        "swapouts": sstat.mem.swouts,
    }
    yield values


def parse_PRC(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PRC' parseable representing process cpu usage."""
    for stat in tstats:
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "pid": stat.gen.pid,
            "name": stat.gen.name.decode(),
            "state": stat.gen.state.decode(),
            "ticks": header.hertz,
            "user_consumption": stat.cpu.utime,
            "system_consumption": stat.cpu.stime,
            "nice": stat.cpu.nice,
            "priority": stat.cpu.prio,
            "priority_realtime": stat.cpu.rtprio,
            "policy": stat.cpu.policy,
            "cpu": stat.cpu.curcpu,
            "sleep": stat.cpu.sleepavg,
        }
        yield values


def parse_PRD(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PRD' parseable representing process drive usage."""
    for stat in tstats:
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "pid": stat.gen.pid,
            "name": stat.gen.name.decode(),
            "state": stat.gen.state.decode(),
            "kernel_patch": "y" if header.supportflags & atop_1_26.PATCHSTAT else "n",
            "standard_io": "y" if header.supportflags & atop_1_26.IOSTAT else "n",
            "reads": stat.dsk.rio,
            "read_sectors": stat.dsk.rsz,
            "writes": stat.dsk.wio,
            "written_sectors": stat.dsk.wsz,
            "cancelled_sector_writes": stat.dsk.cwsz,
        }
        yield values


def parse_PRG(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PRG' parseable representing process generic details."""
    for stat in tstats:
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "pid": stat.gen.pid,
            "name": stat.gen.name.decode(),
            "state": stat.gen.state.decode(),
            "real_uid": stat.gen.ruid,
            "real_gid": stat.gen.rgid,
            "tgid": stat.gen.pid,  # This is a duplicate of pid per atop documentation.
            "threads": stat.gen.nthr,
            "exit_code": stat.gen.excode,
            "start_time": stat.gen.btime,
            "cmd": stat.gen.cmdline.decode(),
            "ppid": stat.gen.ppid,
            "running_threads": stat.gen.nthrrun,
            "sleeping_threads": stat.gen.nthrslpi,
            "dead_threads": stat.gen.nthrslpu,
            "effective_uid": stat.gen.euid,
            "effective_gid": stat.gen.egid,
            "saved_uid": stat.gen.suid,
            "saved_gid": stat.gen.sgid,
            "filesystem_uid": stat.gen.fsuid,
            "filesystem_gid": stat.gen.fsgid,
            "elapsed_time": stat.gen.elaps,
        }
        yield values


def parse_PRM(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PRM' parseable representing process memory usage."""
    for stat in tstats:
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "pid": stat.gen.pid,
            "name": stat.gen.name.decode(),
            "state": stat.gen.state.decode(),
            "page": header.pagesize,
            "vsize": stat.mem.vmem * 1024,
            "rsize": stat.mem.rmem * 1024,
            "ssize": stat.mem.shtext * 1024,
            "vgrowth": stat.mem.vgrow * 1024,
            "rgrowth": stat.mem.rgrow * 1024,
            "minor_faults": stat.mem.minflt,
            "major_faults": stat.mem.majflt,
        }
        yield values


def parse_PRN(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'PRN' parseable representing process network activity."""
    for stat in tstats:
        values = {
            "timestamp": record.curtime,
            "interval": record.interval,
            "pid": stat.gen.pid,
            "name": stat.gen.name.decode(),
            "state": stat.gen.state.decode(),
            "kernel_patch": "y" if header.supportflags & atop_1_26.PATCHSTAT else "n",
            "tcp_transmitted": stat.net.tcpsnd,
            "tcp_transmitted_size": stat.net.tcpssz,
            "tcp_received": stat.net.tcprcv,
            "tcp_received_size": stat.net.tcprsz,
            "udp_transmitted": stat.net.udpsnd,
            "udp_transmitted_size": stat.net.udpssz,
            "udp_received": stat.net.udprcv,
            "udp_received_size": stat.net.udprsz,
            "raw_transmitted": stat.net.rawsnd,
            "raw_received": stat.net.rawrcv,
        }
        yield values


def parse_SWP(
    header: utils.Header,
    record: utils.Record,
    sstat: utils.SStat,
    tstats: list[utils.TStat],
) -> dict:
    """Retrieves statistics for Atop 'SWP' parseable representing swap space usage."""
    values = {
        "timestamp": record.curtime,
        "interval": record.interval,
        "page_size": header.pagesize,
        "swap": sstat.mem.totswap,
        "free_swap": sstat.mem.freeswap,
        "committed_space": sstat.mem.committed,
        "committed_limit": sstat.mem.commitlim,
    }
    yield values



"""Structs and definitions used serialize/deserialize Atop statistics directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h

Struct ordering matches the C source to help with comparisons.

See https://github.com/Atoptool/atop for more information and full details about each field.
Using schemas and structs from Atop 1.26.
"""

import ctypes

from atoparser.structs.shared import HeaderMixin
from atoparser.structs.shared import UTSName
from atoparser.structs.shared import count_t
from atoparser.structs.shared import time_t

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name,too-few-public-methods

# Definitions from atop.h
ACCTACTIVE = 0x00000001
PATCHSTAT = 0x00000002
IOSTAT = 0x00000004
PATCHACCT = 0x00000008

# Definitions from photoproc.h
PNAMLEN = 15
CMDLEN = 150

# Definitions from photosyst.h
MAXCPU = 64
MAXDSK = 256
MAXLVM = 256
MAXMDD = 128
MAXDKNAM = 32
MAXINTF = 32


class Record(ctypes.Structure):
    """Top level struct to describe basic process information, and the following SStat and PStat structs.

    C Name: rawrecord
    C Location: rawlog.c
    """

    _fields_ = [
        ("curtime", time_t),
        ("flags", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 3),
        ("scomplen", ctypes.c_uint),
        ("pcomplen", ctypes.c_uint),
        ("interval", ctypes.c_uint),
        ("nlist", ctypes.c_uint),
        ("npresent", ctypes.c_uint),
        ("nexit", ctypes.c_uint),
        ("ntrun", ctypes.c_uint),
        ("ntslpi", ctypes.c_uint),
        ("ntslpu", ctypes.c_uint),
        ("nzombie", ctypes.c_uint),
        ("ifuture", ctypes.c_uint * 6),
    ]


class MemStat(ctypes.Structure):
    """Embedded struct to describe basic memory information.

    C Name: memstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("physmem", count_t),
        ("freemem", count_t),
        ("buffermem", count_t),
        ("slabmem", count_t),
        ("cachemem", count_t),
        ("cachedrt", count_t),
        ("totswap", count_t),
        ("freeswap", count_t),
        ("pgscans", count_t),
        ("allocstall", count_t),
        ("swouts", count_t),
        ("swins", count_t),
        ("commitlim", count_t),
        ("committed", count_t),
        ("cfuture", count_t * 4),
    ]


class FreqCnt(ctypes.Structure):
    """Embedded struct to describe basic processor frequency information.

    C Name: freqcnt
    C Location: photosyst.h
    C Parent: percpu
    """

    _fields_ = [
        ("maxfreq", count_t),
        ("cnt", count_t),
        ("ticks", count_t),
    ]


class PerCPU(ctypes.Structure):
    """Embedded struct to describe per processor usage information.

    C Name: percpu
    C Location: photosyst.h
    C Parent: cpustat
    """

    _fields_ = [
        ("cpunr", ctypes.c_int),
        ("stime", count_t),
        ("utime", count_t),
        ("ntime", count_t),
        ("itime", count_t),
        ("wtime", count_t),
        ("Itime", count_t),
        ("Stime", count_t),
        ("steal", count_t),
        ("guest", count_t),
        ("freqcnt", FreqCnt),
        ("cfuture", count_t * 1),
    ]


class CPUStat(ctypes.Structure):
    """Embedded struct to describe basic overall processor information.

    C Name: cpustat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrcpu", count_t),
        ("devint", count_t),
        ("csw", count_t),
        ("nprocs", count_t),
        ("lavg1", ctypes.c_float),
        ("lavg5", ctypes.c_float),
        ("lavg15", ctypes.c_float),
        ("cfuture", count_t * 4),
        ("all", PerCPU),
        ("cpu", PerCPU * MAXCPU),
    ]
    fields_limiters = {"cpu": "nrcpu"}


class IPv4Stats(ctypes.Structure):
    """Embedded struct to describe overall IPv4 statistics.

    C Name: ipv4_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("Forwarding", count_t),
        ("DefaultTTL", count_t),
        ("InReceives", count_t),
        ("InHdrErrors", count_t),
        ("InAddrErrors", count_t),
        ("ForwDatagrams", count_t),
        ("InUnknownProtos", count_t),
        ("InDiscards", count_t),
        ("InDelivers", count_t),
        ("OutRequests", count_t),
        ("OutDiscards", count_t),
        ("OutNoRoutes", count_t),
        ("ReasmTimeout", count_t),
        ("ReasmReqds", count_t),
        ("ReasmOKs", count_t),
        ("ReasmFails", count_t),
        ("FragOKs", count_t),
        ("FragFails", count_t),
        ("FragCreates", count_t),
    ]


class ICMPv4Stats(ctypes.Structure):
    """Embedded struct to describe overall ICMPv4 statistics.

    C Name: icmpv4_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("InMsgs", count_t),
        ("InErrors", count_t),
        ("InDestUnreachs", count_t),
        ("InTimeExcds", count_t),
        ("InParmProbs", count_t),
        ("InSrcQuenchs", count_t),
        ("InRedirects", count_t),
        ("InEchos", count_t),
        ("InEchoReps", count_t),
        ("InTimestamps", count_t),
        ("InTimestampReps", count_t),
        ("InAddrMasks", count_t),
        ("InAddrMaskReps", count_t),
        ("OutMsgs", count_t),
        ("OutErrors", count_t),
        ("OutDestUnreachs", count_t),
        ("OutTimeExcds", count_t),
        ("OutParmProbs", count_t),
        ("OutSrcQuenchs", count_t),
        ("OutRedirects", count_t),
        ("OutEchos", count_t),
        ("OutEchoReps", count_t),
        ("OutTimestamps", count_t),
        ("OutTimestampReps", count_t),
        ("OutAddrMasks", count_t),
        ("OutAddrMaskReps", count_t),
    ]


class UDPv4Stats(ctypes.Structure):
    """Embedded struct to describe overall UDPv4 statistics.

    C Name: udpv4_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("InDatagrams", count_t),
        ("NoPorts", count_t),
        ("InErrors", count_t),
        ("OutDatagrams", count_t),
    ]


class TCPStats(ctypes.Structure):
    """Embedded struct to describe overall TCP statistics.

    C Name: tcp_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("RtoAlgorithm", count_t),
        ("RtoMin", count_t),
        ("RtoMax", count_t),
        ("MaxConn", count_t),
        ("ActiveOpens", count_t),
        ("PassiveOpens", count_t),
        ("AttemptFails", count_t),
        ("EstabResets", count_t),
        ("CurrEstab", count_t),
        ("InSegs", count_t),
        ("OutSegs", count_t),
        ("RetransSegs", count_t),
        ("InErrs", count_t),
        ("OutRsts", count_t),
    ]


class IPv6Stats(ctypes.Structure):
    """Embedded struct to describe overall IPv6 statistics.

    C Name: ipv6_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("Ip6InReceives", count_t),
        ("Ip6InHdrErrors", count_t),
        ("Ip6InTooBigErrors", count_t),
        ("Ip6InNoRoutes", count_t),
        ("Ip6InAddrErrors", count_t),
        ("Ip6InUnknownProtos", count_t),
        ("Ip6InTruncatedPkts", count_t),
        ("Ip6InDiscards", count_t),
        ("Ip6InDelivers", count_t),
        ("Ip6OutForwDatagrams", count_t),
        ("Ip6OutRequests", count_t),
        ("Ip6OutDiscards", count_t),
        ("Ip6OutNoRoutes", count_t),
        ("Ip6ReasmTimeout", count_t),
        ("Ip6ReasmReqds", count_t),
        ("Ip6ReasmOKs", count_t),
        ("Ip6ReasmFails", count_t),
        ("Ip6FragOKs", count_t),
        ("Ip6FragFails", count_t),
        ("Ip6FragCreates", count_t),
        ("Ip6InMcastPkts", count_t),
        ("Ip6OutMcastPkts", count_t),
    ]


class ICMPv6Stats(ctypes.Structure):
    """Embedded struct to describe overall ICMPv6 statistics.

    C Name: icmpv6_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("Icmp6InMsgs", count_t),
        ("Icmp6InErrors", count_t),
        ("Icmp6InDestUnreachs", count_t),
        ("Icmp6InPktTooBigs", count_t),
        ("Icmp6InTimeExcds", count_t),
        ("Icmp6InParmProblems", count_t),
        ("Icmp6InEchos", count_t),
        ("Icmp6InEchoReplies", count_t),
        ("Icmp6InGroupMembQueries", count_t),
        ("Icmp6InGroupMembResponses", count_t),
        ("Icmp6InGroupMembReductions", count_t),
        ("Icmp6InRouterSolicits", count_t),
        ("Icmp6InRouterAdvertisements", count_t),
        ("Icmp6InNeighborSolicits", count_t),
        ("Icmp6InNeighborAdvertisements", count_t),
        ("Icmp6InRedirects", count_t),
        ("Icmp6OutMsgs", count_t),
        ("Icmp6OutDestUnreachs", count_t),
        ("Icmp6OutPktTooBigs", count_t),
        ("Icmp6OutTimeExcds", count_t),
        ("Icmp6OutParmProblems", count_t),
        ("Icmp6OutEchoReplies", count_t),
        ("Icmp6OutRouterSolicits", count_t),
        ("Icmp6OutNeighborSolicits", count_t),
        ("Icmp6OutNeighborAdvertisements", count_t),
        ("Icmp6OutRedirects", count_t),
        ("Icmp6OutGroupMembResponses", count_t),
        ("Icmp6OutGroupMembReductions", count_t),
    ]


class UDPv6Stats(ctypes.Structure):
    """Embedded struct to describe overall UDPv6 statistics.

    C Name: udpv6_stats
    C Location: netstats.h
    C Parent: netstat
    """

    _fields_ = [
        ("Udp6InDatagrams", count_t),
        ("Udp6NoPorts", count_t),
        ("Udp6InErrors", count_t),
        ("Udp6OutDatagrams", count_t),
    ]


class NETStat(ctypes.Structure):
    """Embedded struct to describe overall network statistics.

    C Name: netstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("ipv4", IPv4Stats),
        ("icmpv4", ICMPv4Stats),
        ("udpv4", UDPv4Stats),
        ("ipv6", IPv6Stats),
        ("icmpv6", ICMPv6Stats),
        ("udpv6", UDPv6Stats),
        ("tcp", TCPStats),
    ]


class PerDSK(ctypes.Structure):
    """Embedded struct to describe per disk information.

    C Name: perdsk
    C Location: photosyst.h
    C Parent: dskstat
    """

    _fields_ = [
        ("name", ctypes.c_char * MAXDKNAM),
        ("nread", count_t),
        ("nrsect", count_t),
        ("nwrite", count_t),
        ("nwsect", count_t),
        ("io_ms", count_t),
        ("avque", count_t),
        ("cfuture", count_t * 4),
    ]


class DSKStat(ctypes.Structure):
    """Embedded struct to describe overall disk information.

    C Name: dskstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("ndsk", ctypes.c_int),
        ("nmdd", ctypes.c_int),
        ("nlvm", ctypes.c_int),
        ("dsk", PerDSK * MAXDSK),
        ("mdd", PerDSK * MAXMDD),
        ("lvm", PerDSK * MAXLVM),
    ]
    fields_limiters = {
        "dsk": "ndsk",
        "mdd": "nmdd",
        "lvm": "nlvm",
    }


class PerIntf(ctypes.Structure):
    """Embedded struct to describe per interface statistics.

    C Name: perintf
    C Location: photosyst.h
    C Parent: intfstat
    """

    _fields_ = [
        ("name", ctypes.c_char * 16),
        ("rbyte", count_t),
        ("rpack", count_t),
        ("rerrs", count_t),
        ("rdrop", count_t),
        ("rfifo", count_t),
        ("rframe", count_t),
        ("rcompr", count_t),
        ("rmultic", count_t),
        ("rfuture", count_t * 4),
        ("sbyte", count_t),
        ("spack", count_t),
        ("serrs", count_t),
        ("sdrop", count_t),
        ("sfifo", count_t),
        ("scollis", count_t),
        ("scarrier", count_t),
        ("scompr", count_t),
        ("sfuture", count_t * 4),
        ("speed", ctypes.c_long),
        ("duplex", ctypes.c_char),
        ("cfuture", count_t * 4),
    ]


class IntfStat(ctypes.Structure):
    """Embedded struct to describe overall interface statistics.

    C Name: intfstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrintf", ctypes.c_int),
        ("intf", PerIntf * MAXINTF),
    ]
    fields_limiters = {
        "intf": "nrintf",
    }


class WWWStat(ctypes.Structure):
    """Embedded struct to describe experimental statistics for local HTTP daemons.

    C Name: wwwstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("accesses", count_t),
        ("totkbytes", count_t),
        ("uptime", count_t),
        ("bworkers", ctypes.c_int),
        ("iworkers", ctypes.c_int),
    ]


class SStat(ctypes.Structure):
    """Top level struct to describe various subsystems.

    C Name: sstat
    C Location: photosyst.h
    """

    _fields_ = [
        ("cpu", CPUStat),
        ("mem", MemStat),
        ("net", NETStat),
        ("intf", IntfStat),
        ("dsk", DSKStat),
        ("www", WWWStat),
    ]


class GEN(ctypes.Structure):
    """Embedded struct to describe a single process' general information.

    C Name: gen
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("pid", ctypes.c_int),
        ("ppid", ctypes.c_int),
        ("ruid", ctypes.c_int),
        ("euid", ctypes.c_int),
        ("suid", ctypes.c_int),
        ("fsuid", ctypes.c_int),
        ("rgid", ctypes.c_int),
        ("egid", ctypes.c_int),
        ("sgid", ctypes.c_int),
        ("fsgid", ctypes.c_int),
        ("nthr", ctypes.c_int),
        ("name", ctypes.c_char * (PNAMLEN + 1)),
        ("state", ctypes.c_char),
        ("excode", ctypes.c_int),
        ("btime", time_t),
        ("elaps", time_t),
        ("cmdline", ctypes.c_char * (CMDLEN + 1)),
        ("nthrslpi", ctypes.c_int),
        ("nthrslpu", ctypes.c_int),
        ("nthrrun", ctypes.c_int),
        ("ifuture", ctypes.c_int * 1),
    ]


class CPU(ctypes.Structure):
    """Embedded struct to describe a single process' processor usage.

    C Name: cpu
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("utime", count_t),
        ("stime", count_t),
        ("nice", ctypes.c_int),
        ("prio", ctypes.c_int),
        ("rtprio", ctypes.c_int),
        ("policy", ctypes.c_int),
        ("curcpu", ctypes.c_int),
        ("sleepavg", ctypes.c_int),
        ("ifuture", ctypes.c_int * 4),
        ("cfuture", count_t * 4),
    ]


class DSK(ctypes.Structure):
    """Embedded struct to describe a single process' disk usage.

    C Name: dsk
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("rio", count_t),
        ("rsz", count_t),
        ("wio", count_t),
        ("wsz", count_t),
        ("cwsz", count_t),
        ("cfuture", count_t * 4),
    ]


class MEM(ctypes.Structure):
    """Embedded struct to describe a single process' memory usage.

    C Name: mem
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("minflt", count_t),
        ("majflt", count_t),
        ("shtext", count_t),
        ("vmem", count_t),
        ("rmem", count_t),
        ("vgrow", count_t),
        ("rgrow", count_t),
        ("cfuture", count_t * 4),
    ]


class NET(ctypes.Structure):
    """Embedded struct to describe a single process' network usage.

    C Name: net
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("tcpsnd", count_t),
        ("tcpssz", count_t),
        ("tcprcv", count_t),
        ("tcprsz", count_t),
        ("udpsnd", count_t),
        ("udpssz", count_t),
        ("udprcv", count_t),
        ("udprsz", count_t),
        ("rawsnd", count_t),
        ("rawrcv", count_t),
        ("cfuture", count_t * 4),
    ]


class PStat(ctypes.Structure):
    """Top level struct to describe multiple statistic categories per process.

    C Name: pstat
    C Location: photoproc.h
    """

    _fields_ = [
        ("gen", GEN),
        ("cpu", CPU),
        ("dsk", DSK),
        ("mem", MEM),
        ("net", NET),
    ]


# Add TStat alias for forwards compatibility. TStat and PStat are the same base stats, but renamed in later versions.
TStat = PStat


class Header(ctypes.Structure, HeaderMixin):
    """Top level struct to describe information about the system running Atop and the log file itself.

    C Name: rawheader
    C Location: rawlog.c
    """

    _fields_ = [
        ("magic", ctypes.c_uint),
        ("aversion", ctypes.c_ushort),
        ("future1", ctypes.c_ushort),
        ("future2", ctypes.c_ushort),
        ("rawheadlen", ctypes.c_ushort),
        ("rawreclen", ctypes.c_ushort),
        ("hertz", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 6),
        ("sstatlen", ctypes.c_uint),
        ("pstatlen", ctypes.c_uint),
        ("utsname", UTSName),
        ("cfuture", ctypes.c_char * 8),
        ("pagesize", ctypes.c_uint),
        ("supportflags", ctypes.c_int),
        ("osrel", ctypes.c_int),
        ("osvers", ctypes.c_int),
        ("ossub", ctypes.c_int),
        ("ifuture", ctypes.c_int * 6),
    ]
    supported_version = "1.26"
    Record = Record
    SStat = SStat
    PStat = PStat
    # Add TStat alias for forwards compatibility. TStat and PStat are the same base stats, but renamed in later versions.
    TStat = PStat
    CStat = None
    CGChainer = None



"""Structs and definitions used serialize/deserialize Atop statistics directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h

Struct ordering matches the C source to help with comparisons.
If structs match exactly from a previous version, they are reused via aliasing.

See https://github.com/Atoptool/atop for more information and full details about each field.
Using schemas and structs from Atop 2.4.0.
"""

import ctypes

from atoparser.structs import atop_1_26
from atoparser.structs import atop_2_3
from atoparser.structs.shared import count_t

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name,too-few-public-methods

# Definitions from atop.h
ACCTACTIVE = 0x00000001
IOSTAT = 0x00000004
NETATOP = 0x00000010
NETATOPD = 0x00000020
DOCKSTAT = 0x00000040
GPUSTAT = 0x00000080

# Definitions from photoproc.h
PNAMLEN = 15
CMDLEN = 255

# Definitions from photosyst.h
MAXCPU = 2048
MAXDSK = 1024
MAXLVM = 2048
MAXMDD = 256
MAXINTF = 128
MAXCONTAINER = 128
MAXNFSMOUNT = 64
MAXIBPORT = 32
MAXGPU = 32
MAXGPUBUS = 12
MAXGPUTYPE = 12
MAXDKNAM = 32
MAXIBNAME = 12


Record = atop_2_3.Record


MemStat = atop_2_3.MemStat


FreqCnt = atop_1_26.FreqCnt


class PerCPU(ctypes.Structure):
    """Embedded struct to describe per processor usage information.

    C Name: percpu
    C Location: photosyst.h
    C Parent: cpustat
    """

    _fields_ = [
        ("cpunr", ctypes.c_int),
        ("stime", count_t),
        ("utime", count_t),
        ("ntime", count_t),
        ("itime", count_t),
        ("wtime", count_t),
        ("Itime", count_t),
        ("Stime", count_t),
        ("steal", count_t),
        ("guest", count_t),
        ("freqcnt", FreqCnt),
        ("instr", count_t),
        ("cycle", count_t),
        ("cfuture", count_t * 2),
    ]


class CPUStat(ctypes.Structure):
    """Embedded struct to describe basic overall processor information.

    C Name: cpustat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrcpu", count_t),
        ("devint", count_t),
        ("csw", count_t),
        ("nprocs", count_t),
        ("lavg1", ctypes.c_float),
        ("lavg5", ctypes.c_float),
        ("lavg15", ctypes.c_float),
        ("cfuture", count_t * 4),
        ("all", PerCPU),
        ("cpu", PerCPU * MAXCPU),
    ]
    fields_limiters = {"cpu": "nrcpu"}


PerDSK = atop_2_3.PerDSK


DSKStat = atop_2_3.DSKStat


PerIntf = atop_2_3.PerIntf


IntfStat = atop_2_3.IntfStat


PerNFSMount = atop_2_3.PerNFSMount


Server = atop_2_3.Server


Client = atop_2_3.Client


NFSMounts = atop_2_3.NFSMounts


NFSStat = atop_2_3.NFSStat


class PSI(ctypes.Structure):
    """Embedded struct to describe average pressure.

    C Name: psi
    C Location: photosyst.h
    C Parent: pressure
    """

    _fields_ = [
        ("avg10", ctypes.c_float),
        ("avg60", ctypes.c_float),
        ("avg300", ctypes.c_float),
        ("total", count_t),
    ]


class Pressure(ctypes.Structure):
    """Embedded struct to describe pressure stats.

    C Name: pressure
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("present", ctypes.c_char),
        ("future", ctypes.c_char * 3),
        ("cpusome", PSI),
        ("memsome", PSI),
        ("memfull", PSI),
        ("iosome", PSI),
        ("iofull", PSI),
    ]


PerContainer = atop_2_3.PerContainer


ContStat = atop_2_3.ContStat


WWWStat = atop_1_26.WWWStat


class PerGPU(ctypes.Structure):
    """Embedded struct to describe per GPU statistics.

    C Name: pergpu
    C Location: photosyst.h
    C Parent: gpustat
    """

    _fields_ = [
        ("taskstats", ctypes.c_char),
        ("nrprocs", ctypes.c_uint8),
        ("type", ctypes.c_char * (MAXGPUTYPE + 1)),
        ("busid", ctypes.c_char * (MAXGPUBUS + 1)),
        ("gpunr", ctypes.c_int),
        ("gpupercnow", ctypes.c_int),
        ("mempercnow", ctypes.c_int),
        ("memtotnow", count_t),
        ("memusenow", count_t),
        ("samples", count_t),
        ("gpuperccum", count_t),
        ("memperccum", count_t),
        ("memusecum", count_t),
    ]


class GPUStat(ctypes.Structure):
    """Embedded struct to describe overall GPU statistics.

    C Name: gpustat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrgpus", ctypes.c_int),
        ("gpu", PerGPU * MAXGPU),
    ]
    fields_limiters = {
        "gpu": "nrgpus",
    }


class PerIFB(ctypes.Structure):
    """Embedded struct to describe per InfiniBand statistics.

    C Name: perifb
    C Location: photosyst.h
    C Parent: ifbstat
    """

    _fields_ = [
        ("ibname", ctypes.c_char * MAXIBNAME),
        ("portnr", ctypes.c_short),
        ("lanes", ctypes.c_short),
        ("rate", count_t),
        ("rcvb", count_t),
        ("sndb", count_t),
        ("rcvp", count_t),
        ("sndp", count_t),
    ]


class IFBStat(ctypes.Structure):
    """Embedded struct to describe overall InfiniBand statistics.

    C Name: ifbstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrports", ctypes.c_int),
        ("ifb", PerIFB * MAXIBPORT),
    ]
    fields_limiters = {
        "ifb": "nrports",
    }


IPv4Stats = atop_1_26.IPv4Stats


ICMPv4Stats = atop_1_26.ICMPv4Stats


UDPv4Stats = atop_1_26.UDPv4Stats


TCPStats = atop_1_26.TCPStats


IPv6Stats = atop_1_26.IPv6Stats


ICMPv6Stats = atop_1_26.ICMPv6Stats


UDPv6Stats = atop_1_26.UDPv6Stats


NETStat = atop_1_26.NETStat


class SStat(ctypes.Structure):
    """Top level struct to describe various subsystems.

    C Name: sstat
    C Location: photosyst.h
    """

    _fields_ = [
        ("cpu", CPUStat),
        ("mem", MemStat),
        ("net", NETStat),
        ("intf", IntfStat),
        ("dsk", DSKStat),
        ("nfs", NFSStat),
        ("cfs", ContStat),
        ("psi", Pressure),
        ("gpu", GPUStat),
        ("ifb", IFBStat),
        ("www", WWWStat),
    ]


GEN = atop_2_3.GEN


CPU = atop_1_26.CPU


DSK = atop_1_26.DSK


MEM = atop_2_3.MEM


NET = atop_2_3.NET


class GPU(ctypes.Structure):
    """Embedded struct to describe a single process' GPU usage.

    C Name: gpu
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("state", ctypes.c_char),
        ("cfuture", ctypes.c_char * 3),
        ("nrgpus", ctypes.c_short),
        ("gpulist", ctypes.c_int32),
        ("gpubusy", ctypes.c_int),
        ("membusy", ctypes.c_int),
        ("timems", count_t),
        ("memnow", count_t),
        ("memcum", count_t),
        ("sample", count_t),
    ]


class TStat(ctypes.Structure):
    """Top level struct to describe multiple statistic categories per task/process.

    C Name: tstat
    C Location: photoproc.h
    """

    _fields_ = [
        ("gen", GEN),
        ("cpu", CPU),
        ("dsk", DSK),
        ("mem", MEM),
        ("net", NET),
        ("gpu", GPU),
    ]


class Header(atop_2_3.Header):
    """Top level struct to describe information about the system running Atop and the log file itself."""

    supported_version = "2.4"
    Record = Record
    SStat = SStat
    TStat = TStat
    CStat = None
    CGChainer = None



"""Structs and definitions used serialize/deserialize Atop statistics directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h

Struct ordering matches the C source to help with comparisons.
If structs match exactly from a previous version, they are reused via aliasing.

See https://github.com/Atoptool/atop for more information and full details about each field.
Using schemas and structs from Atop 2.3.0.
"""

import ctypes

from atoparser.structs import atop_1_26
from atoparser.structs.shared import HeaderMixin
from atoparser.structs.shared import UTSName
from atoparser.structs.shared import count_t
from atoparser.structs.shared import time_t

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name,too-few-public-methods

# Definitions from atop.h
ACCTACTIVE = 0x00000001
IOSTAT = 0x00000004
NETATOP = 0x00000010
NETATOPD = 0x00000020
DOCKSTAT = 0x00000040

# Definitions from photoproc.h
PNAMLEN = 15
CMDLEN = 255

# Definitions from photosyst.h
MAXCPU = 2048
MAXDSK = 1024
MAXLVM = 2048
MAXMDD = 256
MAXINTF = 128
MAXCONTAINER = 128
MAXNFSMOUNT = 64
MAXDKNAM = 32


class Record(ctypes.Structure):
    """Top level struct to describe basic process information, and the following SStat and TStat structs.

    C Name: rawrecord
    C Location: rawlog.c
    """

    _fields_ = [
        ("curtime", time_t),
        ("flags", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 3),
        ("scomplen", ctypes.c_uint),
        ("pcomplen", ctypes.c_uint),
        ("interval", ctypes.c_uint),
        ("ndeviat", ctypes.c_uint),
        ("nactproc", ctypes.c_uint),
        ("ntask", ctypes.c_uint),
        ("totproc", ctypes.c_uint),
        ("totrun", ctypes.c_uint),
        ("totslpi", ctypes.c_uint),
        ("totslpu", ctypes.c_uint),
        ("totzomb", ctypes.c_uint),
        ("nexit", ctypes.c_uint),
        ("noverflow", ctypes.c_uint),
        ("ifuture", ctypes.c_uint * 6),
    ]


class MemStat(ctypes.Structure):
    """Embedded struct to describe basic memory information.

    C Name: memstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("physmem", count_t),
        ("freemem", count_t),
        ("buffermem", count_t),
        ("slabmem", count_t),
        ("cachemem", count_t),
        ("cachedrt", count_t),
        ("totswap", count_t),
        ("freeswap", count_t),
        ("pgscans", count_t),
        ("pgsteal", count_t),
        ("allocstall", count_t),
        ("swouts", count_t),
        ("swins", count_t),
        ("commitlim", count_t),
        ("committed", count_t),
        ("shmem", count_t),
        ("shmrss", count_t),
        ("shmswp", count_t),
        ("slabreclaim", count_t),
        ("tothugepage", count_t),
        ("freehugepage", count_t),
        ("hugepagesz", count_t),
        ("vmwballoon", count_t),
        ("cfuture", count_t * 8),
    ]


FreqCnt = atop_1_26.FreqCnt


class PerCPU(ctypes.Structure):
    """Embedded struct to describe per processor usage information.

    C Name: percpu
    C Location: photosyst.h
    C Parent: cpustat
    """

    _fields_ = [
        ("cpunr", ctypes.c_int),
        ("stime", count_t),
        ("utime", count_t),
        ("ntime", count_t),
        ("itime", count_t),
        ("wtime", count_t),
        ("Itime", count_t),
        ("Stime", count_t),
        ("steal", count_t),
        ("guest", count_t),
        ("freqcnt", FreqCnt),
        ("cfuture", count_t * 4),
    ]


class CPUStat(ctypes.Structure):
    """Embedded struct to describe basic overall processor information.

    C Name: cpustat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrcpu", count_t),
        ("devint", count_t),
        ("csw", count_t),
        ("nprocs", count_t),
        ("lavg1", ctypes.c_float),
        ("lavg5", ctypes.c_float),
        ("lavg15", ctypes.c_float),
        ("cfuture", count_t * 4),
        ("all", PerCPU),
        ("cpu", PerCPU * MAXCPU),
    ]
    fields_limiters = {"cpu": "nrcpu"}


class PerDSK(ctypes.Structure):
    """Embedded struct to describe per disk information.

    C Name: perdsk
    C Location: photosyst.h
    C Parent: dskstat
    """

    _fields_ = [
        ("name", ctypes.c_char * MAXDKNAM),
        ("nread", count_t),
        ("nrsect", count_t),
        ("nwrite", count_t),
        ("nwsect", count_t),
        ("io_ms", count_t),
        ("avque", count_t),
        ("cfuture", count_t * 4),
    ]


class DSKStat(ctypes.Structure):
    """Embedded struct to describe overall disk information.

    C Name: dskstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("ndsk", ctypes.c_int),
        ("nmdd", ctypes.c_int),
        ("nlvm", ctypes.c_int),
        ("dsk", PerDSK * MAXDSK),
        ("mdd", PerDSK * MAXMDD),
        ("lvm", PerDSK * MAXLVM),
    ]
    fields_limiters = {
        "dsk": "ndsk",
        "mdd": "nmdd",
        "lvm": "nlvm",
    }


class PerIntf(ctypes.Structure):
    """Embedded struct to describe per interface statistics.

    C Name: perintf
    C Location: photosyst.h
    C Parent: intfstat
    """

    _fields_ = [
        ("name", ctypes.c_char * 16),
        ("rbyte", count_t),
        ("rpack", count_t),
        ("rerrs", count_t),
        ("rdrop", count_t),
        ("rfifo", count_t),
        ("rframe", count_t),
        ("rcompr", count_t),
        ("rmultic", count_t),
        ("rfuture", count_t * 4),
        ("sbyte", count_t),
        ("spack", count_t),
        ("serrs", count_t),
        ("sdrop", count_t),
        ("sfifo", count_t),
        ("scollis", count_t),
        ("scarrier", count_t),
        ("scompr", count_t),
        ("sfuture", count_t * 4),
        ("type", ctypes.c_char),
        ("speed", ctypes.c_long),
        ("speedp", ctypes.c_long),
        ("duplex", ctypes.c_char),
        ("cfuture", count_t * 4),
    ]


class IntfStat(ctypes.Structure):
    """Embedded struct to describe overall interface statistics.

    C Name: intfstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrintf", ctypes.c_int),
        ("intf", PerIntf * MAXINTF),
    ]
    fields_limiters = {
        "intf": "nrintf",
    }


class PerNFSMount(ctypes.Structure):
    """Embedded struct to describe per NFS mount statistics.

    C Name: pernfsmount
    C Location: photosyst.h
    C Parent: nfsmounts
    """

    _fields_ = [
        ("mountdev", ctypes.c_char * 128),
        ("age", count_t),
        ("bytesread", count_t),
        ("byteswrite", count_t),
        ("bytesdread", count_t),
        ("bytesdwrite", count_t),
        ("bytestotread", count_t),
        ("bytestotwrite", count_t),
        ("pagesmread", count_t),
        ("pagesmwrite", count_t),
        ("future", count_t * 8),
    ]


class Server(ctypes.Structure):
    """Embedded struct to describe NFS server information.

    C Name: server
    C Location: photoproc.h
    C Parent: nfsstat
    """

    _fields_ = [
        ("netcnt", count_t),
        ("netudpcnt", count_t),
        ("nettcpcnt", count_t),
        ("nettcpcon", count_t),
        ("rpccnt", count_t),
        ("rpcbadfmt", count_t),
        ("rpcbadaut", count_t),
        ("rpcbadcln", count_t),
        ("rpcread", count_t),
        ("rpcwrite", count_t),
        ("rchits", count_t),
        ("rcmiss", count_t),
        ("rcnoca", count_t),
        ("nrbytes", count_t),
        ("nwbytes", count_t),
        ("future", count_t * 8),
    ]


class Client(ctypes.Structure):
    """Embedded struct to describe NFS client information.

    C Name: client
    C Location: photoproc.h
    C Parent: nfsstat
    """

    _fields_ = [
        ("rpccnt", count_t),
        ("rpcretrans", count_t),
        ("rpcautrefresh", count_t),
        ("rpcread", count_t),
        ("rpcwrite", count_t),
        ("future", count_t * 8),
    ]


class NFSMounts(ctypes.Structure):
    """Embedded struct to describe NFS mount information.

    C Name: mfsmounts
    C Location: photoproc.h
    C Parent: nfsstat
    """

    _fields_ = [
        ("nrmounts", ctypes.c_int),
        ("nfsmnt", PerNFSMount * MAXNFSMOUNT),
    ]
    fields_limiters = {
        "nfsmnt": "nrmounts",
    }


class NFSStat(ctypes.Structure):
    """Embedded struct to describe NFS subsystem.

    C Name: nfstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("server", Server),
        ("client", Client),
        ("nfsmounts", NFSMounts),
    ]


class PerContainer(ctypes.Structure):
    """Embedded struct to describe per container statistics.

    C Name: percontainer
    C Location: photosyst.h
    C Parent: constat
    """

    _fields_ = [
        ("ctid", ctypes.c_ulong),
        ("numproc", ctypes.c_ulong),
        ("system", count_t),
        ("user", count_t),
        ("nice", count_t),
        ("uptime", count_t),
        ("physpages", count_t),
    ]


class ContStat(ctypes.Structure):
    """Embedded struct to describe container subsystem.

    C Name: contstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrcontainer", ctypes.c_int),
        ("cont", PerContainer * MAXCONTAINER),
    ]
    fields_limiters = {
        "cont": "nrcontainer",
    }


WWWStat = atop_1_26.WWWStat


IPv4Stats = atop_1_26.IPv4Stats


ICMPv4Stats = atop_1_26.ICMPv4Stats


UDPv4Stats = atop_1_26.UDPv4Stats


TCPStats = atop_1_26.TCPStats


IPv6Stats = atop_1_26.IPv6Stats


ICMPv6Stats = atop_1_26.ICMPv6Stats


UDPv6Stats = atop_1_26.UDPv6Stats


NETStat = atop_1_26.NETStat


class SStat(ctypes.Structure):
    """Top level struct to describe various subsystems.

    C Name: sstat
    C Location: photosyst.h
    """

    _fields_ = [
        ("cpu", CPUStat),
        ("mem", MemStat),
        ("net", NETStat),
        ("intf", IntfStat),
        ("dsk", DSKStat),
        ("nfs", NFSStat),
        ("cfs", ContStat),
        ("www", WWWStat),
    ]


class GEN(ctypes.Structure):
    """Embedded struct to describe a single process' general information.

    C Name: gen
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("tgid", ctypes.c_int),
        ("pid", ctypes.c_int),
        ("ppid", ctypes.c_int),
        ("ruid", ctypes.c_int),
        ("euid", ctypes.c_int),
        ("suid", ctypes.c_int),
        ("fsuid", ctypes.c_int),
        ("rgid", ctypes.c_int),
        ("egid", ctypes.c_int),
        ("sgid", ctypes.c_int),
        ("fsgid", ctypes.c_int),
        ("nthr", ctypes.c_int),
        ("name", ctypes.c_char * (PNAMLEN + 1)),
        ("isproc", ctypes.c_char),
        ("state", ctypes.c_char),
        ("excode", ctypes.c_int),
        ("btime", time_t),
        ("elaps", time_t),
        ("cmdline", ctypes.c_char * (CMDLEN + 1)),
        ("nthrslpi", ctypes.c_int),
        ("nthrslpu", ctypes.c_int),
        ("nthrrun", ctypes.c_int),
        ("ctid", ctypes.c_int),
        ("vpid", ctypes.c_int),
        ("wasinactive", ctypes.c_int),
        ("container", ctypes.c_char * 16),
    ]


CPU = atop_1_26.CPU


DSK = atop_1_26.DSK


class MEM(ctypes.Structure):
    """Embedded struct to describe a single process' memory usage.

    C Name: mem
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("minflt", count_t),
        ("majflt", count_t),
        ("vexec", count_t),
        ("vmem", count_t),
        ("rmem", count_t),
        ("pmem", count_t),
        ("vgrow", count_t),
        ("rgrow", count_t),
        ("vdata", count_t),
        ("vstack", count_t),
        ("vlibs", count_t),
        ("vswap", count_t),
        ("cfuture", count_t * 4),
    ]


class NET(ctypes.Structure):
    """Embedded struct to describe a single process' network usage.

    C Name: net
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("tcpsnd", count_t),
        ("tcpssz", count_t),
        ("tcprcv", count_t),
        ("tcprsz", count_t),
        ("udpsnd", count_t),
        ("udpssz", count_t),
        ("udprcv", count_t),
        ("udprsz", count_t),
        ("avail1", count_t),
        ("avail2", count_t),
        ("cfuture", count_t * 4),
    ]


class TStat(ctypes.Structure):
    """Top level struct to describe multiple statistic categories per task/process.

    C Name: tstat
    C Location: photoproc.h
    """

    _fields_ = [
        ("gen", GEN),
        ("cpu", CPU),
        ("dsk", DSK),
        ("mem", MEM),
        ("net", NET),
    ]


class Header(ctypes.Structure, HeaderMixin):
    """Top level struct to describe information about the system running Atop and the log file itself.

    C Name: rawheader
    C Location: rawlog.c
    """

    _fields_ = [
        ("magic", ctypes.c_uint),
        ("aversion", ctypes.c_ushort),
        ("future1", ctypes.c_ushort),
        ("future2", ctypes.c_ushort),
        ("rawheadlen", ctypes.c_ushort),
        ("rawreclen", ctypes.c_ushort),
        ("hertz", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 6),
        ("sstatlen", ctypes.c_uint),
        ("tstatlen", ctypes.c_uint),
        ("utsname", UTSName),
        ("cfuture", ctypes.c_char * 8),
        ("pagesize", ctypes.c_uint),
        ("supportflags", ctypes.c_int),
        ("osrel", ctypes.c_int),
        ("osvers", ctypes.c_int),
        ("ossub", ctypes.c_int),
        ("ifuture", ctypes.c_int * 6),
    ]
    supported_version = "2.3"
    Record = Record
    SStat = SStat
    TStat = TStat
    CStat = None
    CGChainer = None



#! /usr/bin/env python3

"""Simple Atop log processor."""

import argparse
import gzip
import json

import atoparser
from atoparser.parsers import atop_1_26

PARSEABLES = ["cpu", "CPL", "CPU", "DSK", "LVM", "MDD", "MEM", "NETL", "NETU", "PAG", "PRC", "PRG", "PRM", "PRN", "SWP"]
PARSEABLE_MAP = {
    "1.26": {parseable: getattr(atop_1_26, f"parse_{parseable}") for parseable in PARSEABLES},
}


def parse_args() -> argparse.Namespace:
    """Parse user arguments.

    Returns:
        Namespace with all the user arguments.
    """
    parser = argparse.ArgumentParser(
        "Convert binary files into JSON output.",
    )
    parser.add_argument(
        "files",
        nargs="+",
        help="Files to process. May be uncompressed or gzip compressed.",
    )
    parser.add_argument(
        "-P",
        "--parseables",
        nargs="+",
        choices=PARSEABLES,
        metavar="PARSEABLE",
        help='Display output in Atop "parseable" format, instead of full structs. See "man atop" for full details. e.g. "CPU", "DSK", etc.',
    )
    parser.add_argument(
        "-p",
        "--pretty-print",
        action="store_true",
        help="Pretty print the JSON output with indentation.",
    )
    parser.add_argument(
        "--tstats",
        action="store_true",
        help="Include TStats/PStats in output. Very verbose.",
    )
    parser.add_argument(
        "--cstats",
        action="store_true",
        help="Include CGroup/CStats in output. Only available with Atop 2.11+ logs. Verbose.",
    )
    args = parser.parse_args()
    return args


def main() -> None:
    """Primary function to load Atop data."""
    args = parse_args()

    for file in args.files:
        samples = []
        opener = open if ".gz" not in file else gzip.open
        with opener(file, "rb") as raw_file:
            header = atoparser.get_header(raw_file)
            if args.parseables and header.semantic_version not in PARSEABLE_MAP:
                samples.append(
                    {
                        "error": f"Atop version {header.semantic_version} does not support parseables, only full raw output.",
                        "file": file,
                    }
                )
                continue
            parsers = PARSEABLE_MAP.get(header.semantic_version, PARSEABLE_MAP["1.26"])
            for record, sstat, tstats, cgroups in atoparser.generate_statistics(
                raw_file,
                header,
                raise_on_truncation=False,
            ):
                if args.parseables:
                    for parseable in args.parseables:
                        for sample in parsers[parseable](header, record, sstat, tstats):
                            sample["parseable"] = parseable
                            samples.append(sample)
                else:
                    converted = {
                        "header": atoparser.struct_to_dict(header),
                        "record": atoparser.struct_to_dict(record),
                        "sstat": atoparser.struct_to_dict(sstat),
                    }
                    if args.tstats:
                        converted["tstat"] = [atoparser.struct_to_dict(stat) for stat in tstats]
                    if args.cstats:
                        converted["cgroup"] = [atoparser.struct_to_dict(stat) for stat in cgroups]
                    samples.append(converted)
        print(json.dumps(samples, indent=2 if args.pretty_print else None))


if __name__ == "__main__":
    main()



"""Libraries for reading Atop raw data files."""

from atoparser.utils import CGChainer
from atoparser.utils import CStat
from atoparser.utils import Header
from atoparser.utils import Record
from atoparser.utils import SStat
from atoparser.utils import TStat
from atoparser.utils import generate_statistics
from atoparser.utils import get_cstat
from atoparser.utils import get_header
from atoparser.utils import get_record
from atoparser.utils import get_sstat
from atoparser.utils import get_tstat
from atoparser.utils import struct_to_dict

__version__ = "3.3.1"



"""Shared structs and definitions used serialize/deserialize Atop data directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h
"""

import ctypes

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name

# Definitions from time.h
time_t = ctypes.c_long

# Definitions from atop.h
count_t = ctypes.c_longlong

# Definitions from sys/types.h
off_t = ctypes.c_long
pid_t = ctypes.c_int


class HeaderMixin:
    """Shared logic for top level struct describing information contained in the log file.

    Attributes:
        supported_version: The version of Atop that this header is compatible with as <major.<minor>.
    """

    supported_version: str
    Record: ctypes.Structure
    SStat: ctypes.Structure
    TStat: ctypes.Structure
    CStat: ctypes.Structure
    CGChainer: ctypes.Structure

    def check_compatibility(self) -> None:
        """Verify if the loaded values are compatible with this header version.

        Raises:
            ValueError if not compatible.
        """
        sizes = [
            ("Header", self.rawheadlen, ctypes.sizeof(self.__class__)),
            ("Record", self.rawreclen, ctypes.sizeof(self.Record)),
            ("SStat", self.sstatlen, ctypes.sizeof(self.SStat)),
        ]
        if self.major_version >= 2 and self.minor_version >= 3:
            sizes.append(("TStat", self.tstatlen, ctypes.sizeof(self.TStat)))
        else:
            sizes.append(("PStat", self.pstatlen, ctypes.sizeof(self.TStat)))
        if self.major_version >= 2 and self.minor_version >= 11:
            sizes.append(("CStat", self.cstatlen, ctypes.sizeof(self.CStat)))
        if any(size[1] != size[2] for size in sizes):
            raise ValueError(
                f"File has incompatible Atop format. Struct length evaluations (type, found, expected): {sizes}"
            )

    @property
    def major_version(self) -> int:
        """The major version from the semantic version."""
        self.semantic_version  # Call the primary property to ensure population. pylint: disable=pointless-statement
        return self._major_version

    @property
    def minor_version(self) -> int:
        """The minor version from the semantic version."""
        self.semantic_version  # Call the primary property to ensure population. pylint: disable=pointless-statement
        return self._minor_version

    @property
    def semantic_version(self) -> str:
        """The semantic version as major.minor.

        Atop releases have "maintenance" versions, but they do not impact the header or file structure.
        i.e., 2.3.1 is the same as 2.3.
        """
        # Use a general getattr() call to ensure the instance can always set the attribute even on first call.
        # C structs have various ways of creating instances, so __init__ is not always called to set up attributes.
        if not getattr(self, "_semantic_version", None):
            # pylint: disable=attribute-defined-outside-init
            self._major_version = (self.aversion >> 8) & 0x7F
            self._minor_version = self.aversion & 0xFF
            self._semantic_version = f"{self._major_version}.{self._minor_version}"
        return self._semantic_version


class UTSName(ctypes.Structure):
    """Struct to describe basic system information.

    C Name: utsname
    C Location: sys/utsname.h
    """

    _fields_ = [
        # Standard GNU length is 64 characters + null terminator
        ("sysname", ctypes.c_char * 65),
        ("nodename", ctypes.c_char * 65),
        ("release", ctypes.c_char * 65),
        ("version", ctypes.c_char * 65),
        ("machine", ctypes.c_char * 65),
        ("domain", ctypes.c_char * 65),
    ]



"""Structs and definitions used serialize/deserialize Atop statistics directly from log files.

Structs are declared in a way that will help provide as close to a 1 to 1 match as possible for debuggability
and maintenance. The _fields_ of every struct match their original name, however the struct names have been updated
to match python CamelCase standards. Each struct includes the following to help identify the original source:
    C Name: utsname
    C Location: sys/utsname.h

Struct ordering matches the C source to help with comparisons.
If structs match exactly from a previous version, they are reused via aliasing.

See https://github.com/Atoptool/atop for more information and full details about each field.
Using schemas and structs from Atop 2.11.0.
"""

import ctypes

from atoparser.structs import atop_1_26
from atoparser.structs import atop_2_3
from atoparser.structs import atop_2_4
from atoparser.structs import atop_2_7
from atoparser.structs import atop_2_8
from atoparser.structs import atop_2_10
from atoparser.structs.shared import HeaderMixin
from atoparser.structs.shared import UTSName
from atoparser.structs.shared import count_t
from atoparser.structs.shared import pid_t
from atoparser.structs.shared import time_t

# Disable the following pylint warnings to allow the variables and classes to match the style from the C.
# This helps with maintainability and cross-referencing.
# pylint: disable=invalid-name,too-few-public-methods

# Definitions from atop.h
ACCTACTIVE = 0x00000001
IOSTAT = 0x00000004
NETATOP = 0x00000010
NETATOPD = 0x00000020
CONTAINERSTAT = 0x00000040
GPUSTAT = 0x00000080
CGROUPV2 = 0x00000100
NETATOPBPF = 0x00001000

# Definitions from photoproc.h
PNAMLEN = 15
CMDLEN = 255
CGRLEN = 64
UTSLEN = 15

# Definitions from photosyst.h
MAXCPU = 2048
MAXDSK = 1024
MAXNUMA = 1024
MAXLVM = 2048
MAXMDD = 256
MAXINTF = 128
MAXCONTAINER = 128
MAXNFSMOUNT = 64
MAXIBPORT = 32
MAXGPU = 32
MAXGPUBUS = 12
MAXGPUTYPE = 12
MAXLLC = 256
MAXDKNAM = 32
MAXIBNAME = 12


class Record(ctypes.Structure):
    """Top level struct to describe basic process information, and the following SStat and TStat structs.

    C Name: rawrecord
    C Location: rawlog.h
    """

    _fields_ = [
        ("curtime", time_t),
        ("flags", ctypes.c_ushort),
        ("ncgroups", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 2),
        ("scomplen", ctypes.c_uint),
        ("pcomplen", ctypes.c_uint),
        ("interval", ctypes.c_uint),
        ("ndeviat", ctypes.c_uint),
        ("nactproc", ctypes.c_uint),
        ("ntask", ctypes.c_uint),
        ("totproc", ctypes.c_uint),
        ("totrun", ctypes.c_uint),
        ("totslpi", ctypes.c_uint),
        ("totslpu", ctypes.c_uint),
        ("totzomb", ctypes.c_uint),
        ("nexit", ctypes.c_uint),
        ("noverflow", ctypes.c_uint),
        ("totidle", ctypes.c_uint),
        ("ccomplen", ctypes.c_uint),
        ("coriglen", ctypes.c_uint),
        ("ncgpids", ctypes.c_uint),
        ("icomplen", ctypes.c_uint),
        ("ifuture", ctypes.c_uint),
    ]


class MemStat(ctypes.Structure):
    """Embedded struct to describe basic memory information.

    C Name: memstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("physmem", count_t),
        ("freemem", count_t),
        ("buffermem", count_t),
        ("slabmem", count_t),
        ("cachemem", count_t),
        ("cachedrt", count_t),
        ("totswap", count_t),
        ("freeswap", count_t),
        ("pgscans", count_t),
        ("pgsteal", count_t),
        ("allocstall", count_t),
        ("swouts", count_t),
        ("swins", count_t),
        ("tcpsock", count_t),
        ("udpsock", count_t),
        ("commitlim", count_t),
        ("committed", count_t),
        ("shmem", count_t),
        ("shmrss", count_t),
        ("shmswp", count_t),
        ("slabreclaim", count_t),
        ("stothugepage", count_t),
        ("sfreehugepage", count_t),
        ("shugepagesz", count_t),
        ("vmwballoon", count_t),
        ("zfsarcsize", count_t),
        ("swapcached", count_t),
        ("ksmsharing", count_t),
        ("ksmshared", count_t),
        ("zswapped", count_t),
        ("zswap", count_t),
        ("oomkills", count_t),
        ("compactstall", count_t),
        ("pgmigrate", count_t),
        ("numamigrate", count_t),
        ("pgouts", count_t),
        ("pgins", count_t),
        ("pagetables", count_t),
        ("zswouts", count_t),
        ("zswins", count_t),
        ("ltothugepage", count_t),
        ("lfreehugepage", count_t),
        ("lhugepagesz", count_t),
        ("availablemem", count_t),
        ("anonhugepage", count_t),
        ("cfuture", count_t * 5),
    ]


class MemPerNUMA(ctypes.Structure):
    """Embedded struct to describe basic memory information per NUMA node.

    C Name: mempernuma
    C Location: photosyst.h
    C Parent: memnuma
    """

    _fields_ = [
        ("numanr", ctypes.c_int),
        ("frag", ctypes.c_float),
        ("totmem", count_t),
        ("freemem", count_t),
        ("filepage", count_t),
        ("dirtymem", count_t),
        ("slabmem", count_t),
        ("slabreclaim", count_t),
        ("active", count_t),
        ("inactive", count_t),
        ("shmem", count_t),
        ("tothp", count_t),
        ("freehp", count_t),
        ("cfuture", count_t * 2),
    ]


class MemNUMA(ctypes.Structure):
    """Embedded struct to describe memory usage across all NUMA nodes.

    C Name: memnuma
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrnuma", count_t),
        ("numa", MemPerNUMA * MAXNUMA),
    ]
    fields_limiters = {"numa": "nrnuma"}


class CPUPerNUMA(ctypes.Structure):
    """Embedded struct to describe basic CPU information per NUMA node.

    C Name: cpupernuma
    C Location: photosyst.h
    C Parent: cpunuma
    """

    _fields_ = [
        ("numanr", ctypes.c_int),
        ("nrcpu", count_t),
        ("stime", count_t),
        ("utime", count_t),
        ("ntime", count_t),
        ("itime", count_t),
        ("wtime", count_t),
        ("Itime", count_t),
        ("Stime", count_t),
        ("steal", count_t),
        ("guest", count_t),
        ("cfuture", count_t * 2),
    ]


class CPUNUMA(ctypes.Structure):
    """Embedded struct to describe CPU usage across all NUMA nodes.

    C Name: cpunuma
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrnuma", count_t),
        ("numa", CPUPerNUMA * MAXNUMA),
    ]
    fields_limiters = {"numa": "nrnuma"}


FreqCnt = atop_1_26.FreqCnt


PerCPU = atop_2_7.PerCPU


CPUStat = atop_2_7.CPUStat


PerDSK = atop_2_8.PerDSK


DSKStat = atop_2_8.DSKStat


PerIntf = atop_2_3.PerIntf


IntfStat = atop_2_3.IntfStat


PerNFSMount = atop_2_3.PerNFSMount


Server = atop_2_3.Server


Client = atop_2_3.Client


NFSMounts = atop_2_3.NFSMounts


NFSStat = atop_2_3.NFSStat


PSI = atop_2_4.PSI


Pressure = atop_2_4.Pressure


PerContainer = atop_2_3.PerContainer


ContStat = atop_2_3.ContStat


WWWStat = atop_1_26.WWWStat


PerGPU = atop_2_4.PerGPU


GPUStat = atop_2_4.GPUStat


class PerIFB(ctypes.Structure):
    """Embedded struct to describe per InfiniBand statistics.

    C Name: perifb
    C Location: photosyst.h
    C Parent: ifbstat
    """

    _fields_ = [
        ("ibname", ctypes.c_char * MAXIBNAME),
        ("portnr", ctypes.c_short),
        ("lanes", ctypes.c_short),
        ("rate", count_t),
        ("rcvb", count_t),
        ("sndb", count_t),
        ("rcvp", count_t),
        ("sndp", count_t),
        ("cfuture", count_t * 4),
    ]


class IFBStat(ctypes.Structure):
    """Embedded struct to describe overall InfiniBand statistics.

    C Name: ifbstat
    C Location: photosyst.h
    C Parent: sstat
    """

    _fields_ = [
        ("nrports", ctypes.c_int),
        ("ifb", PerIFB * MAXIBPORT),
    ]
    fields_limiters = {
        "ifb": "nrports",
    }


PerLLC = atop_2_8.PerLLC


LLCStat = atop_2_8.LLCStat


IPv4Stats = atop_1_26.IPv4Stats


ICMPv4Stats = atop_2_8.ICMPv4Stats


UDPv4Stats = atop_1_26.UDPv4Stats


TCPStats = atop_2_10.TCPStats


IPv6Stats = atop_1_26.IPv6Stats


ICMPv6Stats = atop_1_26.ICMPv6Stats


UDPv6Stats = atop_1_26.UDPv6Stats


NETStat = atop_2_10.NETStat


class SStat(ctypes.Structure):
    """Top level struct to describe various subsystems.

    C Name: sstat
    C Location: photosyst.h
    """

    _fields_ = [
        ("cpu", CPUStat),
        ("mem", MemStat),
        ("net", NETStat),
        ("intf", IntfStat),
        ("memnuma", MemNUMA),
        ("cpunuma", CPUNUMA),
        ("dsk", DSKStat),
        ("nfs", NFSStat),
        ("cfs", ContStat),
        ("psi", Pressure),
        ("gpu", GPUStat),
        ("ifb", IFBStat),
        ("llc", LLCStat),
        ("www", WWWStat),
    ]


class GEN(ctypes.Structure):
    """Embedded struct to describe a single process' general information.

    C Name: gen
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("tgid", ctypes.c_int),
        ("pid", ctypes.c_int),
        ("ppid", ctypes.c_int),
        ("ruid", ctypes.c_int),
        ("euid", ctypes.c_int),
        ("suid", ctypes.c_int),
        ("fsuid", ctypes.c_int),
        ("rgid", ctypes.c_int),
        ("egid", ctypes.c_int),
        ("sgid", ctypes.c_int),
        ("fsgid", ctypes.c_int),
        ("nthr", ctypes.c_int),
        ("name", ctypes.c_char * (PNAMLEN + 1)),
        ("isproc", ctypes.c_char),
        ("state", ctypes.c_char),
        ("excode", ctypes.c_int),
        ("btime", time_t),
        ("elaps", time_t),
        ("cmdline", ctypes.c_char * (CMDLEN + 1)),
        ("nthrslpi", ctypes.c_int),
        ("nthrslpu", ctypes.c_int),
        ("nthrrun", ctypes.c_int),
        ("nthridle", ctypes.c_int),
        ("ctid", ctypes.c_int),
        ("vpid", ctypes.c_int),
        ("wasinactive", ctypes.c_int),
        ("utsname", ctypes.c_char * (UTSLEN + 1)),
        ("cgroupix", ctypes.c_int),
        ("ifuture", ctypes.c_int * 4),
    ]


class CPU(ctypes.Structure):
    """Embedded struct to describe a single process' processor usage.

    C Name: cpu
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("utime", count_t),
        ("stime", count_t),
        ("nice", ctypes.c_int),
        ("prio", ctypes.c_int),
        ("rtprio", ctypes.c_int),
        ("policy", ctypes.c_int),
        ("curcpu", ctypes.c_int),
        ("sleepavg", ctypes.c_int),
        ("ifuture", ctypes.c_int * 6),
        ("wchan", ctypes.c_char * 16),
        ("rundelay", count_t),
        ("blkdelay", count_t),
        ("nvcsw", count_t),
        ("nivcsw", count_t),
        ("cfuture", count_t * 3),
    ]


DSK = atop_1_26.DSK


class MEM(ctypes.Structure):
    """Embedded struct to describe a single process' memory usage.

    C Name: mem
    C Location: photoproc.h
    C Parent: pstat
    """

    _fields_ = [
        ("minflt", count_t),
        ("majflt", count_t),
        ("vexec", count_t),
        ("vmem", count_t),
        ("rmem", count_t),
        ("pmem", count_t),
        ("vgrow", count_t),
        ("rgrow", count_t),
        ("vdata", count_t),
        ("vstack", count_t),
        ("vlibs", count_t),
        ("vswap", count_t),
        ("vlock", count_t),
        ("cfuture", count_t * 7),
    ]


NET = atop_2_3.NET


class GPU(ctypes.Structure):
    """Embedded struct to describe a single process' GPU usage.

    C Name: gpu
    C Location: photoproc.h
    C Parent: tstat
    """

    _fields_ = [
        ("state", ctypes.c_char),
        ("bfuture", ctypes.c_char * 3),
        ("nrgpus", ctypes.c_short),
        ("gpulist", ctypes.c_int32),
        ("gpubusy", ctypes.c_int),
        ("membusy", ctypes.c_int),
        ("timems", count_t),
        ("memnow", count_t),
        ("memcum", count_t),
        ("sample", count_t),
        ("cfuture", count_t * 3),
    ]


class TStat(ctypes.Structure):
    """Top level struct to describe multiple statistic categories per task/process.

    C Name: tstat
    C Location: photoproc.h
    """

    _fields_ = [
        ("gen", GEN),
        ("cpu", CPU),
        ("dsk", DSK),
        ("mem", MEM),
        ("net", NET),
        ("gpu", GPU),
    ]


class CGGen(ctypes.Structure):
    """Embedded struct to describe a single cgroup's general information.

    C Name: cggen
    C Location: cgroups.h
    C Parent: cstat
    """

    _fields_ = [
        ("structlen", ctypes.c_int),
        ("sequence", ctypes.c_int),
        ("parentseq", ctypes.c_int),
        ("depth", ctypes.c_int),
        ("nprocs", ctypes.c_int),
        ("procsbelow", ctypes.c_int),
        ("namelen", ctypes.c_int),
        ("fullnamelen", ctypes.c_int),
        ("ifuture", ctypes.c_int * 4),
        ("namehash", ctypes.c_long),
        ("lfuture", ctypes.c_long * 4),
    ]


class CGConf(ctypes.Structure):
    """Embedded struct to describe a single cgroup's configuration.

    C Name: cgconfg
    C Location: cgroups.h
    C Parent: cstat
    """

    _fields_ = [
        ("cpuweight", ctypes.c_int),
        ("cpumax", ctypes.c_int),
        ("memmax", count_t),
        ("swpmax", count_t),
        ("dskweight", ctypes.c_int),
        ("ifuture", ctypes.c_int * 5),
        ("cfuture", count_t * 5),
    ]


class CGCPU(ctypes.Structure):
    """Embedded struct to describe a single cgroup's processor usage.

    C Name: cgcpu
    C Location: cgroups.h
    C Parent: cstat
    """

    _fields_ = [
        ("utime", count_t),
        ("stime", count_t),
        ("somepres", count_t),
        ("fullpres", count_t),
        ("cfuture", count_t * 5),
    ]


class CGMem(ctypes.Structure):
    """Embedded struct to describe a single cgroup's memory usage.

    C Name: cgmem
    C Location: cgroups.h
    C Parent: cstat
    """

    _fields_ = [
        ("current", count_t),
        ("anon", count_t),
        ("file", count_t),
        ("kernel", count_t),
        ("shmem", count_t),
        ("somepres", count_t),
        ("fullpres", count_t),
        ("cfuture", count_t * 5),
    ]


class CGDSK(ctypes.Structure):
    """Embedded struct to describe a single cgroup's disk usage.

    C Name: cgdsk
    C Location: cgroups.h
    C Parent: cstat
    """

    _fields_ = [
        ("rbytes", count_t),
        ("wbytes", count_t),
        ("rios", count_t),
        ("wios", count_t),
        ("somepres", count_t),
        ("fullpres", count_t),
        ("cfuture", count_t * 5),
    ]


class CStat(ctypes.Structure):
    """Top level struct to describe general info and metrics per cgroup.

    C Name: cstat
    C Location: cgroups.h
    """

    _fields_ = [
        ("gen", CGGen),
        ("conf", CGConf),
        ("cpu", CGCPU),
        ("mem", CGMem),
        ("dsk", CGDSK),
        ("cgname", ctypes.c_char * 0),  # Variable length character array. N.B. Currently unsupported.
    ]


class CGChainer:
    """Structure used to track a pidlist with a cgroup.

    Due to the variable length of cgchainer values, this is a custom C struct like object, rather than a true C struct.
    It provides the same basic attributes as a C struct to allow conversion to JSON.

    C Name: cgchainer
    C Location: cgroups.h
    """

    _fields_ = [
        ("cstat", CStat),
        ("proclist", ctypes.Array[pid_t]),
    ]

    def __init__(self, cstat: CStat, proclist: ctypes.Array[pid_t]) -> None:
        """Initialize the starting CGroup chain values.

        Args:
            cstat: The primary cstat information applicable to the processes.
            proclist: Process IDs contained in the cgroup.
        """
        self.cstat = cstat
        self.proclist = proclist


class Header(ctypes.Structure, HeaderMixin):
    """Top level struct to describe information about the system running Atop and the log file itself.

    C Name: rawheader
    C Location: rawlog.h
    """

    _fields_ = [
        ("magic", ctypes.c_uint),
        ("aversion", ctypes.c_ushort),
        ("future1", ctypes.c_ushort),
        ("future2", ctypes.c_ushort),
        ("rawheadlen", ctypes.c_ushort),
        ("rawreclen", ctypes.c_ushort),
        ("hertz", ctypes.c_ushort),
        ("pidwidth", ctypes.c_ushort),
        ("sfuture", ctypes.c_ushort * 5),
        ("sstatlen", ctypes.c_uint),
        ("tstatlen", ctypes.c_uint),
        ("utsname", UTSName),
        ("cfuture", ctypes.c_char * 8),
        ("pagesize", ctypes.c_uint),
        ("supportflags", ctypes.c_int),
        ("osrel", ctypes.c_int),
        ("osvers", ctypes.c_int),
        ("ossub", ctypes.c_int),
        ("cstatlen", ctypes.c_int),
        ("ifuture", ctypes.c_int * 5),
    ]
    supported_version = "2.11"
    Record = Record
    SStat = SStat
    TStat = TStat
    CStat = CStat
    CGChainer = CGChainer
